#include <iostream>
#include <fstream>
using std::cout;
using std::fstream;
using std::ios;

#define COMMENT_LENGTH 42
#define SUB_HEADER_LENGTH 15
#define PROG_HEADER_LENGTH 12 //this is prgm0000xyyy: and the two size bytes
#define TOTAL_HEADER_LENGTH (SUB_HEADER_LENGTH + PROG_HEADER_LENGTH)

int nChecksum = 0;

void putChecksummedChar( fstream& fout, unsigned char cChar );

int main( int argc, char** argv )
{
	if( argc != 4 )
	{
		cout << "bin273P - BIN to 73P Converter\n"
			<< "Usage: bin273p [-s|-u|-p] [infile] [outfile]\n"
			<< "-s:        Include to make output squished.\n"
			<< "-u:        Include to make output unsquished.\n"
			<< "-p:        Include to make output squished and patch size field.\n"
			<< "[infile]:  BIN file to use, not including extension.\n"
			<< "[outfile]: 73P file to create, not including extension.";
		return -1;
	}

	fstream fin, fout;
	char cBuffer[128];
	unsigned char cTemp;
	int nRawSize, nTemp;

	//testing
	//argv[1] = new char(3);
	//strcpy( argv[1], "-p" );
	//argv[2] = new char(5);
	//strcpy( argv[2], "test" );
	//argv[3] = new char(5);
	//strcpy( argv[3], "abc" );

	//Try to open infile
	sprintf( cBuffer, argv[2] );
	strcat( cBuffer, ".bin" );
	fin.open( cBuffer, ios::in | ios::binary );
	if( fin.fail() )
	{
		cout << "Could not find input file!";
		return -1;
	}
	//Get file size
	fin.seekg( 0, ios::end );
	nRawSize = fin.tellg();
	fin.seekg( 0 );

	//Sanity checks
	if( (nRawSize == 0) | (nRawSize > 900) )
	{
		cout << "BIN is either empty or larger than 900 bytes!";
		if( argv[1][1] != 'p' )
			return -1;
	}

	//Open outfile
	sprintf( cBuffer, argv[3] );
	strcat( cBuffer, ".73p" );
	fout.open( cBuffer, ios::out | ios::binary );
	if( fout.fail() )
	{
		cout << "Could not create output file!";
		return -1;
	}

	//Write header
	sprintf( cBuffer, "**TI73**" );
	cBuffer[8] = 0x1A;
	cBuffer[9] = 0x0A;
	cBuffer[10] = 0x00;
	fout.write( cBuffer, 11 );

	//Write comment
	for( int i = 0; i < COMMENT_LENGTH; i++ )
		cBuffer[i] = 0;
	fout.write( cBuffer, COMMENT_LENGTH );

	//Write header stuff
	if( (argv[1][1] == 's') | (argv[1][1] == 'p') )
	{
		//Put squished header size
		fout.put( (nRawSize + TOTAL_HEADER_LENGTH) & 0xFF );
		fout.put( ((nRawSize + TOTAL_HEADER_LENGTH) & 0xFF00) >> 8 );

		//Put sub header
		putChecksummedChar( fout, 0x0B );
		putChecksummedChar( fout, 0x00 );
		putChecksummedChar( fout, (nRawSize + PROG_HEADER_LENGTH) & 0xFF );
		putChecksummedChar( fout, ((nRawSize + PROG_HEADER_LENGTH) & 0xFF00) >> 8 );

		//Put variable type and name
		putChecksummedChar( fout, 0x06 ); //protected program
		for( int i = 0; i < strlen(argv[3]); i++ )
			putChecksummedChar( fout, toupper(argv[3][i]) ); //output each character in uppercase
		nTemp = 8 - i;
		for( int i = 0; i < nTemp; i++ )
			putChecksummedChar( fout, 0x00 );

		//Put program size (again)
		putChecksummedChar( fout, (nRawSize + PROG_HEADER_LENGTH) & 0xFF );
		putChecksummedChar( fout, ((nRawSize + PROG_HEADER_LENGTH) & 0xFF00) >> 8 );

		//Put program size bytes
		putChecksummedChar( fout, (nRawSize + PROG_HEADER_LENGTH - 2) & 0xFF );
		putChecksummedChar( fout, ((nRawSize + PROG_HEADER_LENGTH - 2) & 0xFF00) >> 8 );
	}
	else
	{
		//Put unsquished header size
		fout.put( ((2 * nRawSize) + TOTAL_HEADER_LENGTH) & 0xFF );
		fout.put( (((2 * nRawSize) + TOTAL_HEADER_LENGTH) & 0xFF00) >> 8 );

		//Put sub header
		putChecksummedChar( fout, 0x0B );
		putChecksummedChar( fout, 0x00 );
		putChecksummedChar( fout, ((2 * nRawSize) + PROG_HEADER_LENGTH) & 0xFF );
		putChecksummedChar( fout, (((2 * nRawSize) + PROG_HEADER_LENGTH) & 0xFF00) >> 8 );

		//Put variable type and name
		putChecksummedChar( fout, 0x05 ); //protected program
		for( int i = 0; i < strlen(argv[3]); i++ )
			putChecksummedChar( fout, toupper(argv[3][i]) ); //output each character in uppercase
		nTemp = 8 - i;
		for( int i = 0; i < nTemp; i++ )
			putChecksummedChar( fout, 0x00 );

		//Put program size (again)
		putChecksummedChar( fout, ((2 * nRawSize) + PROG_HEADER_LENGTH) & 0xFF );
		putChecksummedChar( fout, (((2 * nRawSize) + PROG_HEADER_LENGTH) & 0xFF00) >> 8 );

		//Put program size bytes
		putChecksummedChar( fout, ((2 * nRawSize) + PROG_HEADER_LENGTH - 2) & 0xFF );
		putChecksummedChar( fout, (((2 * nRawSize) + PROG_HEADER_LENGTH - 2) & 0xFF00) >> 8 );
	}

	//Write assembly header
	cBuffer[0] = 0x5F;
	strcpy( cBuffer + 1, "0000" );
	if( (argv[1][1] == 's') | (argv[1][1] == 'p') )
		cBuffer[5] = '1';
	else
		cBuffer[5] = '0';
	if( argv[1][1] == 'p' )
		sprintf( cBuffer + 6, "%03d", 900 );
	else
		sprintf( cBuffer + 6, "%03d", nRawSize );
	cBuffer[9] = 0x3F;
	for( int i = 0; i < 10; i++ )
		putChecksummedChar( fout, cBuffer[i] );

	//Write program data
	while( !fin.eof() )
	{
		if(( argv[1][1] == 's') | (argv[1][1] == 'p') )
		{
			cTemp = fin.get();
			if( !fin.eof() )
				putChecksummedChar( fout, cTemp );
		}
		else
		{
			cTemp = fin.get();
			if( !fin.eof() )
			{
				sprintf( cBuffer, "%02X", cTemp );
				putChecksummedChar( fout, cBuffer[0] );
				putChecksummedChar( fout, cBuffer[1] );
			}
		}
	}

	//Write checksum
	fout.put( nChecksum & 0xFF );
	fout.put( (nChecksum & 0xFF00) >> 8 );

	//Close files
	fout.close();
	fin.close();

	cout << "Program created successfully!";

	return 0;
}

void putChecksummedChar( fstream& fout, unsigned char cChar )
{
	nChecksum += (int)cChar;
	fout.put( cChar );
}